import threading
from typing import List

from channels.amqp_helpers import Channel, WorkflowEventChannel

from workspaces.capability.enums import (
    CapabilityEventType,
    CapabilityStepType,
    ExecutionPriority,
    ExecutionState,
)
from workspaces.capability.helpers import CapabilitySequence, Parameter
from workspaces.capability.schema import CapabilityEvent, CapabilityRequest
from workspaces.capability.schema_interfaces import (
    CapabilityExecutionIF,
    CapabilityRequestIF,
)
from workspaces.capability.services.capability_queue import CapabilityQueue
from workspaces.capability.services.interfaces import (
    CapabilityInfoIF,
    CapabilityServiceIF,
)
from workspaces.products.schema import FutureProduct
from workspaces.workflow.enum import WorkflowEventType
from workspaces.workflow.schema import WorkflowEvent
from workspaces.workflow.services.interfaces import WorkflowServiceIF


class CapabilityService(CapabilityServiceIF):
    """
    The capability service: clients access this to request capability runs
    """

    def __init__(
        self, capability_info: CapabilityInfoIF, workflow_service: WorkflowServiceIF
    ):
        self.execution_pool = []
        self.queues = {}
        self.capability_info = capability_info
        self.workflow_service = workflow_service
        self.listener = threading.Thread(target=self.listen_for_events, daemon=True)
        self.listener.start()

    def create_request(
        self,
        capability_name: str,
        parameters: List[Parameter] = None,
        products: List[FutureProduct] = None,
    ) -> CapabilityRequestIF:
        """
        Create a capability request for a given capability

        :param capability_name:
        :param parameters: Parameters for request
        :param products: Products the request will create upon success
        :return: Capability request
        """
        return self.capability_info.create_capability_request(
            capability_name, parameters, products
        )

    def run_capability(self, request: CapabilityRequest) -> CapabilityExecutionIF:
        """
        Run a capability with the requested settings

        :param request: Capability request
        :return: Execution of the request
        """
        execution = self.capability_info.create_execution(request)
        steps = CapabilitySequence.from_str(execution.steps)

        if steps[0].step_type == CapabilityStepType.PrepareAndRunWorkflow:
            # First step is to run workflow, move to capability queue
            # FIXME: Priority needs to be dynamic
            self.enqueue_execution(execution, ExecutionPriority.Default.value)
        else:
            self.execution_pool.append(execution)
        return execution

    def update_execution(self, event: WorkflowEvent):
        """
        Update capability execution given a received event
        TODO: Implement execution cancellation

        :param event: Incoming event
        """
        capability_event = self.workflow_to_capability_event(event)

        if not capability_event:
            # WorkflowEvent does not need to update the execution
            return

        execution = self.capability_info.lookup_execution(capability_event.execution_id)
        step_sequence = CapabilitySequence.from_str(execution.steps)
        current_step = step_sequence[execution.current_step]
        event_type = CapabilityEventType[capability_event.event_type]

        # Check to make sure event type is correct
        if current_step.step_type.value == event_type.value:
            # Check if previous step (that just completed) is run workflow step
            if (
                step_sequence[execution.current_step].step_type
                == CapabilityStepType.PrepareAndRunWorkflow.name
            ):
                # Return capability engine to available state
                self.queues[execution.capability.id].complete_execution(
                    capability_event.execution_id
                )

            if execution.current_step != len(step_sequence):
                # Execution is not on its last step
                execution.current_step += 1
            else:
                # Execution on its last step and received the proper event
                execution.state = ExecutionState.Complete.name
                self.capability_info.save_execution(execution)
                return

            # Check if upcoming step is run workflow step
            if (
                step_sequence[execution.current_step].step_type
                == CapabilityStepType.PrepareAndRunWorkflow
            ):
                # Update execution record's state
                execution.state = ExecutionState.ExecutingStep.name
                # Enqueue execution that is on a run workflow step
                # FIXME: Priority needs to be dynamic; perhaps a priority column in execution schema?
                self.enqueue_execution(execution, ExecutionPriority.Default.value)

            # Update execution record's state
            execution.state = ExecutionState.Ready.name
            self.capability_info.save_execution(execution)
        else:
            # TODO: Logging
            print(
                f"Mismatched event type {event_type} for execution {capability_event.execution_id}"
            )

    def enqueue_execution(
        self,
        execution: CapabilityExecutionIF,
        priority: int = ExecutionPriority.Default.value,
    ):
        """
        Move execution record that is ready to execute a workflow into the appropriate capability
        queue

        :param execution: Capability execution record
        :param priority: Priority of the requested execution
        """
        # Get correct queue or initialize one
        queue = self.queues.get(
            execution.capability.id,
            CapabilityQueue(
                self.capability_info,
                self.workflow_service,
                execution.capability.max_jobs,
            ),
        )
        queue.enqueue(execution, priority)
        self.queues[execution.capability.id] = queue

        # Remove execution record from pool
        if execution in self.execution_pool:
            self.execution_pool.remove(execution)

    def workflow_to_capability_event(self, event: WorkflowEvent) -> CapabilityEvent:
        """
        Convert a workflow event into a capability event

        :param event: Workflow event
        :return: Capability event translated from given workflow event
        """
        if event.type == WorkflowEventType.SUBMITTED.name:
            # Workflow submitted
            event_type = CapabilityEventType.WorkflowReady
        elif event.type == WorkflowEventType.TERMINATED.name and event.retval == 0:
            # Workflow completed successfully
            event_type = CapabilityEventType.WorkflowComplete
        else:
            # No corresponding capability event
            return None

        execution = self.capability_info.lookup_execution_by_workflow_request_id(
            event.workflow_request_id
        )

        return CapabilityEvent(
            event_type, execution.capability_request.id, execution.id
        )

    def listen_for_events(self):
        """
        Method that will run in a different thread that will be listening for capability events that it can use
        to update capability executions
        """
        thread_workflow_events = Channel(WorkflowEventChannel(), threaded=True)
        thread_workflow_events.listen(callback=self.update_execution)
