from __future__ import annotations

import json
from pathlib import Path
from typing import Dict

import chevron
import sqlalchemy as sa
from pycapo import CapoConfig
from sqlalchemy import create_engine
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import relationship, sessionmaker

Base = declarative_base()


class AbstractFile:
    """
    Abstract file is exactly that, an abstract concept of what a file is, to be
    returned from various non-filesystem places.
    """

    def __init__(self, filename: str, content: bytes):
        self.filename, self.content = filename, content

    def write_to(self, directory: Path):
        (directory / self.filename).write_bytes(self.content)

    def __json__(self, request):
        return {"filename": self.filename}

    @classmethod
    def from_path(cls, path: Path) -> AbstractFile:
        return cls(path.name, path.read_bytes())


class AbstractRequest:
    """
    Abstract File is an abstract concept that is used to create requests
    """

    name: str
    argument: json

    def __init__(self, name: str, argument: json):
        self.name, self.argument = name, argument

    def __json__(self, request):
        return {"workflow_name": self.name, "argument": self.argument}


class NotificationTemplate(Base):
    """
    Notification Templates are Mustache-formatted files used for sending emails with the notification system.
    """

    __tablename__ = "notification_templates"
    name = sa.Column("name", sa.String, primary_key=True)
    description = sa.Column("description", sa.String, nullable=False)
    content = sa.Column("template", sa.String, nullable=False)

    def render(self, argument: Dict) -> AbstractFile:
        """
        Render a given Mustache-formatted Workflow Template with the specified arguments.

        :param argument:    argument(s) needed for template render
        :return:            an AbstractFile rendered from mustache template
        """

        # TODO: if template has more than one tag, figure out how pass argument parameter
        #  in as JSON and replace 'data' section accordingly
        args = {
            "template": self.content.decode("ascii"),
            "data": {"arguments": "".join(map(str, argument))},
        }
        contents = chevron.render(**args)
        return contents


def get_engine():
    """
    Generate the SQL Alchemy engine for us, using Capo

    :return: SQLAlchemy engine from database URL
    """

    capo = CapoConfig().settings("metadataDatabase")
    url = capo.jdbcUrl.replace("jdbc:", "").replace(
        "://", f"://{capo.jdbcUsername}:{capo.jdbcPassword}@"
    )
    return create_engine(url)


def get_session_factory(engine):
    """
    Generate the SQL Alchemy session factory for us, using the supplied engine

    :param engine:
    :return: SQLAlchemy engine
    """

    factory = sessionmaker(expire_on_commit=False)
    factory.configure(bind=engine)
    return factory
