import json
from pathlib import Path
from typing import Dict, List

import chevron
import sqlalchemy as sa
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import relationship

from workspaces.system.schema import AbstractFile, AbstractRequest
from workspaces.workflow.enum import WorkflowEventType
from workspaces.workflow.schema_interfaces import (
    WorkflowIF,
    WorkflowRequestFileIF,
    WorkflowRequestIF,
)

Base = declarative_base()


class WorkflowEvent:
    """
    An event from the underlying workflow execution system, which is likely
    to be HTCondor.
    """

    def __init__(
        self,
        workflow_request_id: int,
        job_name: str,
        condor_job_id: str,
        event_type: WorkflowEventType,
        timestamp: str,
        log: str,
        retval: int = None,
    ):
        # NOTE: when you modify the properties here, also update the schema
        # in the .json package.
        self.workflow_request_id = workflow_request_id
        self.job_name = job_name
        self.condor_job_id = str(condor_job_id)
        self.type = event_type
        self.timestamp = timestamp
        self.log = log
        self.retval = retval

    def json(self) -> str:
        d = {
            "workflow_request_id": self.workflow_request_id,
            "job_name": self.job_name,
            "type": str(self.type),
            "timestamp": self.timestamp,
            "return_value": self.retval,
            "log": self.log,
        }
        return json.dumps(d)

    def __str__(self):
        return f"WorkflowEvent with data {self.json()}"

    def __repr__(self):
        return f"<WorkflowEvent {self.__dict__}>"

    def __eq__(self, other):
        return (
            self.job_name == other.job_name
            and self.condor_job_id == other.condor_job_id
            and self.type == other.type
            and self.timestamp == other.timestamp
            and self.log == other.log
            and self.retval == other.retval
        )


class Workflow(Base, WorkflowIF):
    """
    A Workflow is a suite of tasks defined by templates that must be
    executed together with some user-submitted files.
    """

    __tablename__ = "workflows"
    workflow_name = sa.Column("workflow_name", sa.String, primary_key=True)
    templates = relationship("WorkflowTemplate", backref="workflows", lazy="dynamic")
    requests = relationship("WorkflowRequest", backref="workflows")

    def __json__(self, request):
        return {"workflow_name": self.workflow_name, "templates": self.templates}

    def __repr__(self):
        return f"<Workflow workflow_name={self.workflow_name}>"

    def render_templates(
        self, argument: Dict, files: List[Path] = None
    ) -> List[AbstractFile]:
        """
        Render the templates associated with this workflow

        :param argument: the workflow argument JSON
        :param files:    the files to be processed
        :return:         a list of rendered template files
        """
        rendered_files = []

        for f in files:
            rendered_files.append(AbstractFile.from_path(f))

        for template in self.templates:
            rendered_files.append(template.render(argument))

        return rendered_files


class WorkflowTemplate(Base):
    """
    Workflow Templates are Mustache-formatted files associated with a given Workflow
    """

    __tablename__ = "workflow_templates"
    filename = sa.Column("filename", sa.String, primary_key=True)
    content = sa.Column("content", sa.LargeBinary, nullable=False)
    workflow_name = sa.Column(
        "workflow_name",
        sa.String,
        sa.ForeignKey("workflows.workflow_name"),
        primary_key=True,
    )

    def render(self, argument: Dict) -> AbstractFile:
        """
        Render a given Mustache-formatted Workflow Template with the specified arguments.

        :param argument:    argument(s) needed for template render
        :return:            an AbstractFile rendered from mustache template
        """

        # TODO: if template has more than one tag, figure out how pass argument parameter
        #  in as JSON and replace 'data' section accordingly
        args = {
            "template": self.content.decode("ascii"),
            "data": {"arguments": "".join(map(str, argument))},
        }
        contents = chevron.render(**args)
        return AbstractFile(self.filename, contents.encode("ascii"))

    @property
    def template_file(self) -> AbstractFile:
        return AbstractFile(self.filename, self.content)

    @template_file.setter
    def template_file(self, file: AbstractFile):
        self.filename, self.content = file.filename, file.content

    def __json__(self, request):
        # Defer to the actual template file's contents for JSON conversion
        return self.template_file.__json__(request)

    def __repr__(self):
        return f"<WorkflowTemplate filename={self.filename} for workflow={self.workflow_name}>"


class WorkflowRequest(Base, WorkflowRequestIF):
    """
    Workflow Requests invoke a particular workflow with a particular argument and some user-supplied files.
    """

    __tablename__ = "workflow_requests"
    workflow_request_id = sa.Column(
        "workflow_request_id", sa.Integer, primary_key=True, autoincrement=True
    )
    workflow_name = sa.Column(
        "workflow_name", sa.String, sa.ForeignKey("workflows.workflow_name")
    )
    argument = sa.Column("argument", sa.JSON)
    state = sa.Column("state", sa.String)
    results_dir = sa.Column("results_dir", sa.String)
    files = relationship("WorkflowRequestFile", backref="request")

    @property
    def request(self) -> AbstractRequest:
        return AbstractRequest(self.workflow_name, self.argument)

    @request.setter
    def create_request(self, request: AbstractRequest):
        self.workflow_name, self.argument = request.name, request.argument

    def update_status(self, status: str):
        self.state = status

    # TODO: create following fields in table
    def set_start_time(self, time: str):
        self.start_time = time

    def set_end_time(self, time: str):
        self.end_time = time

    def __json__(self, request) -> dict:
        return dict(
            workflow_request_id=self.workflow_request_id,
            workflow_name=self.workflow_name,
            argument=self.argument,
            state=self.state,
            results_dir=self.results_dir,
            files=self.files,
        )

    def __getitem__(self, item):
        return self.files

    def __repr__(self):
        return f"<WorkflowRequest workflow_request_id= {self.workflow_request_id}>"


class WorkflowRequestFile(Base, WorkflowRequestFileIF):
    """
    A Workflow Request File is a file supplied by the user and attached to the request they have submitted.
    """

    __tablename__ = "workflow_request_files"
    workflow_request_id = sa.Column(
        "workflow_request_id",
        sa.Integer,
        sa.ForeignKey("workflow_requests.workflow_request_id"),
        primary_key=True,
    )
    filename = sa.Column("filename", sa.String, primary_key=True)
    content = sa.Column("content", sa.LargeBinary, nullable=False)

    @property
    def file(self) -> AbstractFile:
        return AbstractFile(self.filename, self.content)

    @file.setter
    def file(self, file: AbstractFile):
        self.filename, self.content = file.filename, file.content

    def __repr__(self):
        return f"<WorkflowRequestFile filename={self.filename}>"

    def __json__(self, request):
        return {"filename": self.filename}
