import os
import stat
import subprocess
import threading
from pathlib import Path
from tempfile import mkdtemp
from typing import List

import requests
from channels.amqp_helpers import Channel, WorkflowEventChannel
from pycapo import CapoConfig

from workspaces.system.schema import AbstractFile
from workspaces.workflow.enum import WorkflowEventType, WorkflowRequestState
from workspaces.workflow.schema import (
    WorkflowEvent,
    WorkflowRequest,
    WorkflowRequestFile,
)
from workspaces.workflow.schema_interfaces import WorkflowRequestIF
from workspaces.workflow.services.interfaces import WorkflowInfoIF, WorkflowServiceIF


class WorkflowServiceRESTClient(WorkflowServiceIF):
    def __init__(self):
        self.url = (
            CapoConfig()
            .settings("edu.nrao.archive.workspaces.WorkflowSettings")
            .serviceUrl
        )

    def execute(self, request: WorkflowRequestIF, files: List[AbstractFile]):
        # step 1: create the request
        req = requests.post(
            f"{self.url}/workflows/{request.workflow_name}/requests/create?args={request.argument}"
        ).json()

        # step 2: if necessary, pass the files up for this request
        for file in files:
            requests.post(
                f"{self.url}/workflows/requests/{req['workflow_request_id']}/files/{file.filename}",
                body=file.content,
            )

        # step 3: execute the request
        requests.post(
            f"{self.url}/workflows/requests/{req['workflow_request_id']}/submit"
        )


class WorkflowService(WorkflowServiceIF):
    """
    Executes workflows; should be a freestanding service.
    """

    # The next few things that need to happen here are in response to
    # WorkflowEvents that we receive from wf_monitor as the workflow
    # execution evolves. So we have to set up listening at some point
    # in this class
    def __init__(self, info: WorkflowInfoIF):
        self.info = info

        # Start listening for events from the wf_monitor stream
        self.listener = threading.Thread(
            target=lambda: Channel(WorkflowEventChannel(), threaded=True).listen(
                self.on_workflow_event
            )
        )
        self.listener.start()

    def execute(self, request: WorkflowRequest, files: List[AbstractFile] = None):
        """
        Execute a workflow per the supplied parameters.

        :param request: The Workflow Request to execute
        :param files: additional files to execute with request
        :return:
        """
        # 1. look up workflow, returns workflow
        definition = self.info.lookup_workflow_definition(request.workflow_name)

        # 2. render templates to files, returns list of rendered files
        workflow_files = definition.render_templates(request.argument, files)
        for file in workflow_files:
            self.info.save_file(
                request=request, filename=file.filename, content=file.content
            )

        # 4. prepare files for condor execution
        if not request.results_dir:
            # Create temp results directory if the request doesn't already have one (from a previous execution attempt)
            request.results_dir = str(
                self._prepare_files_for_condor(workflow_files).absolute()
            )
            self.info.save_request(request)

        # 5. execute condor and retrieve log file
        log_file = self._execute_prepared(Path(request.results_dir))

        # 6. start reading the log file (background it)
        subprocess.run(
            ["wf_monitor", log_file, str(request.workflow_request_id)],
            cwd=request.results_dir,
        )

    @staticmethod
    def _prepare_files_for_condor(files: List[WorkflowRequestFile]) -> Path:
        """
        Place the files for Condor into a new temp directory and returns the directory.

        :param files:  a dictionary of filename -> content
        :return:       a Path
        """
        # 1. create a temporary directory
        # TODO: decide where to put temp directory and make capo property
        if not os.path.isdir(str(Path.home()) + "/workspaces_tmp"):
            os.mkdir(str(Path.home()) + "/workspaces_tmp")
        temp_folder = Path(mkdtemp(dir=str(Path.home()) + "/workspaces_tmp"))

        # 2. spool each of the temp files to it
        for file in files:
            (temp_folder / file.filename).write_bytes(file.content)

        # 3. make any scripts in there executable
        for file in temp_folder.glob("*.sh"):
            file.chmod(file.stat().st_mode | stat.S_IEXEC)

        # finished, return folder
        return temp_folder

    @staticmethod
    def _execute_prepared(folder: Path) -> Path:
        """
        Execute HTCondor using the named folder as the source of the files.

        :param folder:  the path to the folder to execute
        :return:        the path to the log file
        """
        print(f"executing on folder {folder}")

        # some file in here should end in .dag; that file is our dagman input
        # TODO: not currently using DAG files for vulture
        dagman = list(folder.glob("*.dag"))[0]
        print(f"dagman file {dagman} exists.")

        # need .condor file for vulture submission
        condor = list(folder.glob("*.condor"))[0]
        print(f"condor file {condor} exists.")

        # ensure the log file exists
        logfile = folder / "condor.log"
        print(f"log file {logfile} exists.")
        logfile.touch()

        # submit
        # print("submitting to condor...")
        # subprocess.run(["condor_submit_dag", str(dagman)], cwd=str(folder.absolute()))

        # vulture is a workaround for testing locally without submitting to condor
        print("submitting to vulture...")
        subprocess.run(
            ["vulture", "job", "execute", str(condor)], cwd=str(folder.absolute())
        )

        # return the logfile
        return logfile

    def on_workflow_event(self, event: WorkflowEvent):
        request = self.info.lookup_workflow_request(event.workflow_request_id)

        status: str
        # 2. update request record with new status
        if event.type == WorkflowEventType.EXECUTING:
            status = WorkflowRequestState.Running.name
        elif event.type == WorkflowEventType.TERMINATED:
            if event.retval == 0:
                status = WorkflowRequestState.Complete.name
            else:
                status = WorkflowRequestState.Failed.name
        else:
            status = WorkflowRequestState.Running.name

        print(
            f"Updating state on workflow request {request.workflow_request_id} to {status}..."
        )
        request.update_status(status)
